package alexsa.data.animation;

import java.util.Hashtable;
import java.util.Vector;

import alexsa.data.tvp.CfgNode;
import alexsa.data.Edge;
import alexsa.data.Node;
import alexsa.data.graphs.ShapeGraph;

public class Analysis {
	public final static int
		STRAIGHT		= 1,
		INTERNAL_CYCLE	= 2,
		EXTERNAL_CYCLE	= 4,
		BRANCHED		= 8,
		CROSS_EDGING	= 16,
		FORWARD_EDGING	= 32;

	public int
		tree_edges = 0,
		cross_edges = 0,
		forward_edges = 0,
		internal_cycles = 0,

		branches = 0;

	// Note: This could be reduced to an int, too.
	public Vector
		external_cycles = null;
	
	public Node
		start = null;
		
	public boolean matches(int i) {
		switch (i) {
			case EXTERNAL_CYCLE:
				return hasExternalCycles();
			case INTERNAL_CYCLE:
				return hasInternalCycles();
			case BRANCHED:
				return hasBranches();
			case STRAIGHT:
				return isStraight();
			case CROSS_EDGING:
				return hasCrossEdges();
			case FORWARD_EDGING:
				return hasForwardEdges();
		}
		return false;
	}

	/*	quality returns, how *good* this analysis is according to the
		given criterion. Higher values are considered better.
		This can for example be the depth for a straight branch,
		the number of external cycles for a tree containing external
		cycles and so on
	*/
	public int quality(int i) {
		switch (i) {
			case EXTERNAL_CYCLE:
				return external_cycles != null ? external_cycles.size() : 0;
			case INTERNAL_CYCLE:
				return internal_cycles;
			case BRANCHED:
				return branches;
			case STRAIGHT:
				return tree_edges;
			case CROSS_EDGING:
				return cross_edges;
			case FORWARD_EDGING:
				return forward_edges;
		}
		return 0;
	}

	public Analysis(Edge t) {
		start = t.source;
		analyze(t.target);
	}
	
	protected void analyze(Node s) {
		if (s == null)
			return;
		int count = s.outDegree();
		if (count > 1)
			branches++;
		for (int i = 0; i < count; i++) {
			Transition t = (Transition)s.getOutEdge(i);
			if (t.valid()) {
				switch (t.type) {
					case Edge.TREE_EDGE:
						tree_edges++;
						analyze(t.target);
						break;
					case Edge.CROSS_EDGE:
						// should I follow this edge? What about cross edges
						// leading to back edges?
						cross_edges++;
						break;
					case Edge.FORWARD_EDGE:
						forward_edges++;
						break;
					case Edge.BACK_EDGE:
						if (internal(t))
							internal_cycles++;
						else {
							if (external_cycles == null)
								external_cycles = new Vector();
							external_cycles.addElement(t);
						}
						break;
				}
			}
		}
	}
	
	protected boolean internal(Edge t) {
		Node p = t.source;
		while (p != null && p != t.target && p != start)
			p = p.parent;
		if (p == null) {
			System.err.println("Failure! Couldn't find origin for cycle from " +
				((ShapeGraph)t.source).identifier + " to " + ((ShapeGraph)t.target).identifier);
			return false;
		} else if (p == t.target && p != start)
			return true;
		else
			return false;
	}
	
	public boolean isStraight() {
		return branches + internal_cycles + cross_edges + forward_edges == 0 &&
			external_cycles == null;
	}
	
	public boolean hasCrossEdges() {
		return cross_edges > 0;
	}
	
	public boolean hasExternalCycles() {
		return external_cycles != null;
	}
	
	public boolean hasInternalCycles() {
		return internal_cycles > 0;
	}

	public boolean hasForwardEdges() {
		return forward_edges > 0;
	}
	
	public boolean hasBranches() {
		return branches > 0;
	}
}
