package alexsa.data.graphs;

import java.awt.Point;

import alexsa.data.Node;

public class GraphNode extends Node implements Cloneable {

// ------------- Constants -----------------------	
	public final static int
		DEFAULT_WIDTH  = 50,
		DEFAULT_HEIGHT = 20;

//-------------- Variables -----------------------
	// x and y specify the nodes center!
	public int
		x,
		y,
		width = DEFAULT_WIDTH,
		height = DEFAULT_HEIGHT;

	public boolean
		active = false;

	public String
		label = null;

	public Attributes
		attributes = null;
	
	public String
		significant_label = null;


// -------------- Constructors ------------------
	public GraphNode() {
	}

	public GraphNode(String s, int x, int y)
	{
		this(s, x, y, GraphSymbols.ELLIPSE);
	}

	public GraphNode(String s, int x, int y, int shp)
	{
		this(null, x, y, DEFAULT_WIDTH, DEFAULT_HEIGHT, s, new Attributes(shp));
	}

	public GraphNode(String name, int x, int y, int w, int h, String label, Attributes a)
	{
		this.name = name;
		this.x = x;
		this.y = y;
		width = w;
		height = h;
		attributes = a;
		int pos = label.indexOf("\\n");
		if (attributes.isRound() && pos >= 0) {
			significant_label = label.substring(0, pos);
			this.label = label.substring(pos + 2);
		} else
			this.label = label;
	}
	
	public Object clone() {
		GraphNode gn = new GraphNode();
		gn.x = x;
		gn.y = y;
		gn.width = width;
		gn.height = height;
		gn.active = active;
		gn.label = label;
		gn.attributes = (Attributes)attributes.clone();
		gn.significant_label = significant_label;
		gn.name = name;
		return gn;
	}

//-------------- Methods -----------------------
	public boolean isDotted() {
		return attributes != null && attributes.isDotted();
	}


	
	public int getShape() {
		if (attributes == null)
			return GraphSymbols.BOX;
		else
			return attributes.shape;
	}

	// does this node contain the given coordinates?
	public boolean contains(int cx, int cy) {
		return
			cx >= x - width / 2 &&
			cx <= x + width / 2 &&
			cy >= y - height / 2 &&
			cy <= y + height / 2;
	}

	// Does this node contain the coordinates of the given point p?
	public boolean contains(java.awt.Point p) {
		return contains(p.x, p.y);
	}
	
	// does this node have the same position as the given node gn?
	public boolean samePosition(GraphNode gn) {
		return x == gn.x && y == gn.y;
	}
	
	/* overlapsWith returns, whether this node overlaps with
		the given node n or not. The calculation is simplified by the
		assumption, that both nodes have a rectangular shape
	*/
	public boolean overlapsWith(GraphNode n) {
		int dx = Math.abs(x - n.x),
			dy = Math.abs(y - n.y);
		return width / 2 + n.width / 2 > dx &&
			height / 2 + n.height / 2 > dy;
	}

	public Point getPointOnRimTowards(GraphNode b, int diff) {
		return getPointOnRimTowards(b.x, b.y, diff);
	}

	public Point getPointOnRimTowards(Point b, int diff) {
		return getPointOnRimTowards(b.x, b.y, diff);
	}

	// assumption: width > 0	
	public Point getPointOnRimTowards(int b_x, int b_y, int diff) {
		// first: determine the angle of the line
		double angle;
		if (x - b_x != 0)
			angle = Math.atan(((double)y - (double)b_y) / ((double)x - (double)b_x));
		else
			angle = y - b_y < 0 ? Math.PI / 2 : -Math.PI / 2;
		if (b_x < x)
			angle = (-Math.PI) + angle;

		// ensure angle to be from [-pi ... pi]
		if (angle < -Math.PI)
			angle += Math.PI * 2.0;
		else if (angle > Math.PI)
			angle -= Math.PI * 2.0;
		
		if (getShape() == GraphSymbols.BOX) {
			// the angle from the center to the
			// north-east corner, 0 <= ne_angle <= PI/2
			double ne_angle = Math.atan((double)height / (double)width);

			Point p = new Point(x, y);
			// determine the side the line intersects with
			if (angle <= ne_angle && angle >= -ne_angle) {
				// EAST
				p.x = x + width / 2;
				p.y = (int)(y + width * Math.sin(angle) / 2);
			} else if (angle > ne_angle && angle < Math.PI - ne_angle) {
				// NORTH
				p.x = (int)(x + height * Math.cos(angle) / 2);
				p.y = y + height / 2;
			} else if (angle >= Math.PI - ne_angle || angle <= -Math.PI + ne_angle) {
				// WEST
				p.x = x - width / 2;
				p.y = (int)(y + width * Math.sin(angle) / 2);
			} else if (angle > -Math.PI - ne_angle && angle < -ne_angle) {
				// SOUTH
				p.x = (int)(x + height * Math.cos(angle) / 2);
				p.y = y - height / 2;
			}
			if (diff != 0) {
				p.x += (int)diff * Math.cos(angle);
				p.y += (int)diff * Math.sin(angle);
			}
			return p;
		} else if (getShape() == GraphSymbols.ELLIPSE) {
			double
				a_sqr = width * width / 4.0, // square of half width
				b_sqr = height * height / 4.0, // square of half height
				temp_cos = Math.cos(angle),

				// c is the distance from the center of the ellipse to edge
				// of the ellipse following a ray starting at the center
				// with the given angle
				c = Math.sqrt((b_sqr) / (1 - (a_sqr - b_sqr) / (a_sqr) * temp_cos * temp_cos)),

				// x and y specify the end point of the arrow on the ellipse
				p_x = x + (c + diff) * Math.cos(angle),
				p_y = y + (c + diff) * (Math.sin(angle));
			return new Point((int)p_x, (int)p_y);
		} else
			return null;
	}
	
	public String toString() {
		return "GraphNode [name: " + name + ", label: " + label +"]";
	}
}
