package alexsa.data.graphs;

import java.awt.Point;
import java.awt.geom.Point2D;
import java.util.Hashtable;
import java.util.Vector;

import alexsa.data.Node;

public class ShapeGraph extends Node implements Cloneable {
	
	// Identifier corresponds to the id written by the TVLA .trace-output
	public int
		identifier = -1;

	public int
		scale_factor;

	// in 1/1000 inch
	public int
		width,
		height;
	
	public Vector
		edges,
		nodes;

	public ShapeGraph() {
		edges = null;
		nodes = null;
		width = 0;
		height = 0;
	}

	public ShapeGraph(int sf, int w, int h, Vector nl, Vector el) {
		scale_factor = sf;
		width = w;
		height = h;
		nodes = nl;
		edges = el;
		combineEdgesWithNodes();
		flipGraphVertically();
		addArrowPoints();
	}
	
	public Object clone() {
		ShapeGraph sg = new ShapeGraph();
		sg.width = width;
		sg.height = height;
		sg.scale_factor = scale_factor;
		sg.identifier = identifier;
		if (edges != null) {
			sg.edges = new Vector(edges.size());
			for (int i = 0; i < edges.size(); i++)
				sg.edges.addElement(((GraphEdge)edges.elementAt(i)).clone());
		}
		if (nodes != null) {
			sg.nodes = new Vector(nodes.size());
			for (int i = 0; i < nodes.size(); i++)
				sg.nodes.addElement(((GraphNode)nodes.elementAt(i)).clone());
		}
		sg.combineEdgesWithNodes();
		return sg;
	}
	
	public void flipGraphVertically() {
		if (nodes != null)
			for (int i = 0; i < nodes.size(); i++) {
				GraphNode n = (GraphNode)nodes.elementAt(i);
				n.y = height - n.y;
			}
		if (edges != null)
			for (int i = 0; i < edges.size(); i++) {
				GraphEdge e = (GraphEdge)edges.elementAt(i);
				for (int j = 0; j < e.points.size(); j++) {
					Point p = (Point)e.points.elementAt(j);
					p.y = height - p.y;
				}
				if (e.label != null && e.label.location != null)
					e.label.location.y = height - e.label.location.y;
			}
	}

	public int getNumberOfNodes() {
		return nodes != null ? nodes.size() : 0;
	}
	public GraphNode getNode(int i) {
		return nodes != null && i >= 0 && i < nodes.size() ?
			(GraphNode)nodes.elementAt(i)
			: null;
	}
	// return a node by its identifier (not necessarily the label)
	public GraphNode getNode(String s) {
		if (nodes == null)
			return null;
		for (int i = nodes.size() - 1; i >= 0; i--)
			if (((GraphNode)nodes.elementAt(i)).name.equals(s))
				return (GraphNode)nodes.elementAt(i);
		return null;
	}

	public int getNumberOfEdges() {
		return edges != null ? edges.size() : 0;
	}
	public GraphEdge getEdge(int i) {
		return edges != null && i >= 0 && i < edges.size() ?
			(GraphEdge)edges.elementAt(i)
			: null;
	}
	
	protected void combineEdgesWithNodes() {
		if (edges == null || nodes == null)
			return;
		Hashtable node_hash = new Hashtable();
		for (int i = 0; i < nodes.size(); i++) {
			GraphNode n = (GraphNode)nodes.elementAt(i);
			node_hash.put(n.name, n);
		}
		for (int i = 0; i < edges.size(); i++) {
			GraphEdge e = (GraphEdge)edges.elementAt(i);
			e.source = (GraphNode)node_hash.get(e.source_name);
			e.source.addOutEdge(e);
			e.target = (GraphNode)node_hash.get(e.target_name);
			e.target.addInEdge(e);
		}
	}
	
	/* The graphs generated by dot do not contain the arrow heads for
		the edges. We therefore add the necessary points ourselves once
		and for all. We also check, whether the edge has been drawn the
		wrong way around (can happen for edges heading north)
	*/
	protected void addArrowPoints() {
		if (edges == null)
			return;

		for (int i = edges.size() - 1; i >= 0; i--) {
			GraphEdge e = (GraphEdge)edges.elementAt(i);
			GraphNode target = (GraphNode)e.target;

			Point
				last_point = (Point)e.points.elementAt(e.n - 1);

			// in some rare cases the line is not drawn from the source to
			// the target node but the other way around. We therefore check,
			// whether the start of the line is closer to the target than
			// the end and switch start and end if need be.
			if (e.source != e.target) {
				Point
					first_point = (Point)e.points.elementAt(0);
				int dist_start_sqr =
						(first_point.x - target.x) * (first_point.x - target.x) +
						(first_point.y - target.y) * (first_point.y - target.y),
					dist_end_sqr =
						(last_point.x - target.x) * (last_point.x - target.x) +
						(last_point.y - target.y) * (last_point.y - target.y);
				if (dist_start_sqr < dist_end_sqr) {
					Object p;
					for (int j = 0; j <= e.n / 2; j++) {
						p = e.points.elementAt(i);
						e.points.setElementAt(e.points.elementAt(e.n - 1 - j), j);
						e.points.setElementAt(p, e.n - 1 - j);
					}
					last_point = (Point)e.points.elementAt(e.n - 1);
				}
			}
			e.points.addElement(target.getPointOnRimTowards(last_point, 0));
		}
	}
	
	/* do some analysis of this shape graph. This is used to give
		a rough textual description of this graph whenever the user
		is supposed to select a graph. The textual representations
		should enable the user to identify the graphs easier than just
		by their integer ID.
		Depending on the analysis, the graph can have a brief (less than
		40 char) or longer textual analysis.
	*/
	protected String
		short_analysis = null,
		long_analysis = null;

	public void analyze() {
		createLongAnalysis();
		createShortAnalysis();
	}
	
	protected void createLongAnalysis() {
		int heap_cells = 0,
			summarized_cells = 0;
		Vector vars = new Vector();
		for (int i = 0; i < nodes.size(); i++) {
			GraphNode n = (GraphNode)nodes.elementAt(i);
			if (n.getShape() == GraphSymbols.BOX)
				vars.addElement(n.label);
			else
				heap_cells++;
			if (n.isDotted())
				summarized_cells++;
		}
		StringBuffer buf = new StringBuffer();
		buf.append("Shapegraph #" + identifier);
		if (vars.size() == 0 && heap_cells == 0)
			buf.append(" is empty");
		else {
			if (vars.size() > 0) {
 				buf.append(" contains the ");
 				if (vars.size() == 1)
					buf.append("variable " + vars.elementAt(0));
				else {
					buf.append(vars.size() + " variables ");
					for (int i = 0; i < vars.size() - 2; i++) {
						buf.append(vars.elementAt(i));
						buf.append(", ");
					}
					buf.append(vars.elementAt(vars.size() - 2) + " and " +
						vars.elementAt(vars.size() - 1));
				}
				buf.append(",\n  ");
			} else
				buf.append(" has no variables,\n  ");
			if (heap_cells > 0) {
				if (heap_cells == 1) {
					if (summarized_cells == 1)
						buf.append("and one summarized heap cell."); 
					else
						buf.append("and one heap cell.");
				} else {
					if (heap_cells - summarized_cells == 0)
						buf.append("and " + printQuantity(summarized_cells, "summarized cell") + '.');
					else {
						buf.append(printQuantity((heap_cells - summarized_cells), "heap cell"));
						buf.append(" and " + printQuantity(summarized_cells, "summarized cell") + '.');
					}
				}
			} else {
				buf.append("and no allocated heap cells.");
			}
		}
		buf.append('\n');
		long_analysis = buf.toString();
	}
	
	/* while the long analysis (see above) creates a statistic, the short
		analysis tries to categorize the shape graph.
	*/
	protected void createShortAnalysis() {
		// first: list of length 1?
		if (oneHeapCellAllocated())
			short_analysis = "One allocated heap cell";
		// second; list of length > 1?
		else if (listLengthGreaterOne())
			short_analysis = "List of length > 1";
	}

	// does this shape graph describe exactly one allocated heap cell?
	protected boolean oneHeapCellAllocated() {
		if (nodes.size() != 2 || edges.size() != 1)
			return false;
		GraphNode
			var = null,
			cell = null,
			one, two;
		one = (GraphNode)nodes.elementAt(0);
		two = (GraphNode)nodes.elementAt(1);
		if (one.getShape() == GraphSymbols.BOX && two.getShape() == GraphSymbols.ELLIPSE) {
			var = one;
			cell = two;
		}
		if (one.getShape() == GraphSymbols.ELLIPSE && two.getShape() == GraphSymbols.BOX) {
			var = two;
			cell = one;
		}
		GraphEdge edge = (GraphEdge)edges.elementAt(0);
		return (var != null && cell != null &&
			edge.source_name.equals(var.name) &&
			edge.target_name.equals(cell.name) &&
			!cell.isDotted() &&
			!edge.isDotted());
	}
	
	/* check: is there exactly one variable (box node), one
		non-summarized heap-cell (solid ellipse) and one
		summarized node (dotted ellipse), plus
		one solid edge from the variable to the start of the list,
		one dotted edge to the rest of the list and one dotted
		edge internal to the rest of the list?
	*/ 
	protected boolean listLengthGreaterOne() {
		if (nodes.size() != 3 || edges.size() != 3)
			return false;
		GraphNode
			var = null,
			start = null,
			rest = null;
		for (int i = 0; i < 3; i++) {
			GraphNode n = (GraphNode)nodes.elementAt(i);
			if (n.getShape() == GraphSymbols.BOX) {
				if (var != null)
					return false;
				else
					var = n;
			} else if (n.getShape() == GraphSymbols.ELLIPSE) {
				if (n.isDotted()) {
					if (rest != null)
						return false;
					else
						rest = n;
				} else {
					if (start != null)
						return false;
					else
						start = n;
				}
			}
		}
		if (var == null || start == null || rest == null)
			return false;

		GraphEdge
			allocation = null,
			further = null,
			internal = null;
		for (int i = 0; i < 3; i++) {
			GraphEdge e = (GraphEdge)edges.elementAt(i);
			if (!e.isDotted() && e.source_name.equals(var.name) &&
				e.target_name.equals(start.name))
			{
				if (allocation != null)
					return false;
				else
					allocation = e;
			} else if (e.isDotted() && e.source_name.equals(start.name) &&
				e.target_name.equals(rest.name))
			{
				if (further != null)
					return false;
				else
					further = e;
			} else if (e.isDotted() && e.source_name.equals(rest.name) &&
				e.target_name.equals(rest.name))
			{
				if (internal != null)
					return false;
				else
					internal = e;
			}
		}
		return allocation != null && further != null && internal != null;
	}

	protected String printQuantity(int i, String suffix) {
		if (i == 0)
			return "no " + suffix;
		else if (i == 1)
			return "one " + suffix;
		else
			return "" + i + ' ' + suffix + "s";
	}
	
	public String getShortAnalysis() {
		return short_analysis;
	}
	
	public String getLongAnalysis() {
		return long_analysis;
	}

	public String toString() {
		return "" + identifier;
	}
}
