package alexsa.data.morphing;

import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

import alexsa.data.animation.Transition;
import alexsa.data.graphs.ShapeGraph;
import alexsa.data.graphs.GraphNode;
import alexsa.data.graphs.GraphSymbols;

public class Morphing {
	private final static boolean
		VERBOSE = true;

	public Vector
		splits			= null,
		merges			= null,
		moves			= null,
		label_changes	= null,
		removed			= null,
		created			= null;

	// if there are unconnectable nodes (which often happens, when
	// TVLA found shape a to be isomorphic to shape b), we call the
	// morphing broken, as we cannot establish a well defined morphing.
	// The application is advised to make no transition
	public boolean
		is_broken = false;

	public Morphing(Transition t) {
		Hashtable
			target_nodes = new Hashtable(),
			new_variables = new Hashtable();
		
		for (int i = ((ShapeGraph)t.target).getNumberOfNodes() - 1; i >= 0; i--) {
			GraphNode gn = (GraphNode)((ShapeGraph)t.target).getNode(i);
			if (gn.significant_label != null)
				target_nodes.put(gn.significant_label, gn);
			else if (gn.attributes.isRectangular())
				new_variables.put(gn.label, gn);
		}
		
		Vector
			source_nodes = new Vector(),
			old_variables = new Vector();
		for (int i = ((ShapeGraph)t.source).getNumberOfNodes() - 1; i >= 0; i--) {
			GraphNode gn = (GraphNode)((ShapeGraph)t.source).getNode(i);
			if (gn.significant_label != null)
				source_nodes.addElement(gn);
			else if (gn.attributes.isRectangular())
				old_variables.addElement(gn);
		}
		
		classifyMorphs(source_nodes, target_nodes, t);
		
		classifyVariables(old_variables, new_variables, t);
	}
	
	/** classifyMorphs figures out, what happens to the nodes when the
		transition to the next shape graph takes place. Possible actions
		are:
			Nothing at all
			Renaming
			Moving
			Splitting
			Merging
			Removal
		Input;
			Vector source_nodes, a Vector containing all non-variable-representing
			nodes of a shape graph
			Hashtable target_nodes with all non-variable-representing nodes
			of a shape graph, with the their TVLA-significant-name as key
	*/	
	protected void classifyMorphs(Vector source_nodes, Hashtable target_nodes,
		Transition t)
	{
		for (int i = 0; i < source_nodes.size(); i++) {
			GraphNode source = (GraphNode)source_nodes.elementAt(i);
			GraphNode target = (GraphNode)target_nodes.get(source.significant_label);
			boolean
				split = false,
				merge = false;
			if (target == null) {
				target = (GraphNode)target_nodes.get(source.significant_label +".0");
				if (target != null) {
					addSplit(source, target, (GraphNode)target_nodes.get(source.significant_label +".1"));
					target_nodes.remove(source.significant_label +".0");
					target_nodes.remove(source.significant_label +".1");
					split = true;
				}
			}
			if (target == null) {
				int j = 0;
				while (j < source_nodes.size() && target == null) {
					String possible_name = '[' + source.significant_label + "," +
						((GraphNode)source_nodes.elementAt(j)).significant_label + ']';
					target = (GraphNode)target_nodes.get(possible_name);
					j++;
				}
				if (target != null) {
					j--;
					addMerge(source, ((GraphNode)source_nodes.elementAt(j)), target);
					source_nodes.setElementAt(source_nodes.lastElement(), j);
					source_nodes.setSize(source_nodes.size() - 1);
					target_nodes.remove(target.significant_label);
					merge = true;
				}
			}

			// no merging or splitting? What about movement or
			// label change?
			if (target != null) {
				// does the label change?
				if (!source.label.equals(target.label))
					addLabelChange(source, target);
				// is the node moved?
				if (!source.samePosition(target) && !split && !merge)
					addMovement(source, target);
				target_nodes.remove(target.significant_label);
			} else {
				addRemoval(source);
//				is_broken = true;
				if (VERBOSE)
					System.out.println("Found no targets for node " + source +
						" from shape graph " + ((ShapeGraph)t.source).identifier +
						" in shape graph " + ((ShapeGraph)t.target).identifier);
			}
		}
		Enumeration e = target_nodes.elements();
		while (e.hasMoreElements()) {
			GraphNode gn = (GraphNode)e.nextElement();
			addCreation(gn);
			if (VERBOSE)
				System.out.println("New creation of node " + gn +
					" from shape graph " + ((ShapeGraph)t.source).identifier +
					" in shape graph " + ((ShapeGraph)t.target).identifier);
		}
	}
	
	protected void classifyVariables(Vector old_variables, Hashtable new_variables,
		Transition t)
	{
		for (int i = 0; i < old_variables.size(); i++) {
			GraphNode
				source = (GraphNode)old_variables.elementAt(i),
				target = (GraphNode)new_variables.get(source.label);
			if (target == null) {
				if (VERBOSE)
					System.out.println("Var removed: " + source.label +
						" between " + ((ShapeGraph)t.source).identifier +
						" and " + ((ShapeGraph)t.target).identifier);
				addRemoval(source);
			} else {
				if (!source.samePosition(target))
					addMovement(source, target);
				new_variables.remove(target.label);
			}
		}
		Enumeration e = new_variables.elements();
		while (e.hasMoreElements())
			addCreation((GraphNode)e.nextElement());
	}

	public void addLabelChange(GraphNode a, GraphNode b) {
		if (label_changes == null)
			label_changes = new Vector();
		label_changes.addElement(new Rename(a, b));
	}

	public void addMovement(GraphNode a, GraphNode b) {
		if (moves == null)
			moves = new Vector();
		moves.addElement(new Move(a, b));
	}

	public void addMerge(GraphNode s1, GraphNode s2, GraphNode t) {
		if (merges == null)
			merges = new Vector();
		merges.addElement(new Merge(s1, s2, t));
	}

	public void addSplit(GraphNode s, GraphNode t1, GraphNode t2) {
		if (splits == null)
			splits = new Vector();
		splits.addElement(new Split(s, t1, t2));
	}
	
	public void addRemoval(GraphNode s) {
		if (removed == null)
			removed = new Vector();
		removed.addElement(s);
	}
	
	public void addCreation(GraphNode s) {
		if (created == null)
			created = new Vector();
		created.addElement(s);
	}
	
}