package alexsa.data.tvp;

import java.util.Vector;

public class Formula extends FileElement implements TvpSymbols {
	public Formula
		formula_a = null,
		formula_b = null,
		formula_c = null;

	public int
		operand = -1;

	public Vector
		var_list = null;

	public Predicate
		predicate = null;

	public String
		var_a = null,
		var_b = null,
		var_c = null,
		var_d = null;

	public Kleene
		value = null;


	public Formula() {
	}

	/* ! f
	   this constructor has two parameters even though it only needs one to
	   make a distinction between this constructor and the copy constructor
	   which traditionally only has exactly one parameter of the same type
	   as the class itself (see below for copy constructor)
	 */
	public Formula(Formula f, boolean not) {
		if (f == null)
			nullParameter("(!formula)", "a formula");
		formula_a = f;
		operand = NOT;
	}

	// a -> b, a <-> b, a | b, a & b
	public Formula(Formula a, int op, Formula b) {
		if (a == null)
			nullParameter("(formula o formula)", "a formula");
		if (b == null)
			nullParameter("(formula o formula)", "a second formula");
		formula_a = a;
		operand = op;
		formula_b = b;
	}

	// a ? b : c
	public Formula(Formula a, Formula b, Formula c) {
		if (a == null || b == null || c == null)
			nullParameter("(formula ? formula : formula)", "exactly three formulas");
		formula_a = a;
		formula_b = b;
		formula_c = c;
		operand = QUESTION;
	}

	// a == b, a != b
	public Formula(String a, String b, int op) {
		if (a == null)
			nullParameter("formula (a ==/!= b)", "a variable");
		if (b == null)
			nullParameter("formula (a ==/!= b)", "a second variable");
		var_a = a;
		var_b = b;
		operand = op;
	}

	// A(x) f, E(x) f
	public Formula(Vector v, Formula f, int op) {
		if (f == null)
			nullParameter("Quantified formula", "a formula");
		var_list = v;
		formula_a = f;
		operand = op;
	}

	// predicates (nullary, unary or binary)
	public Formula(Predicate p, Vector v) {
		if (p == null)
			nullParameter("Predicate formula", "a predicate");
		predicate = p;
		var_list = v;
	}

	// transitive closure on binary predicates,
	// reflexive and transitive closure on binary predicates
	public Formula(Predicate p, String a, String b, int op) {
		if (p == null)
			nullParameter("Closure formula", "a predicate");
		if (a == null)
			nullParameter("Closure formula", "a variable");
		if (b == null)
			nullParameter("Closure formula", "a second variable");
		predicate = p;
		var_a = a;
		var_b = b;
		operand = op;
	}

	// transitive closure on a general binary formula
	public Formula(Formula f, String a, String b, String c, String d) {
		if (f == null)
			nullParameter("Transitive closure formula", "a formula");
		if (a == null || b == null || c == null || d == null)
			nullParameter("Transitive closure formula", "four variables");
		operand = TC;
		formula_a = f;
		var_a = a;
		var_b = b;
		var_c = c;
		var_d = d;
	}

	public Formula(Kleene k) {
		if (k == null)
			nullParameter("Atomic formula", "a kleene value");
		value = k;
		operand = MAYBE;
	}


	/* Copy constructor */
	public Formula(Formula f) {
		formula_a = f.formula_a != null ? new Formula(f.formula_a) : null;
		formula_b = f.formula_b != null ? new Formula(f.formula_b) : null;
		formula_c = f.formula_c != null ? new Formula(f.formula_c) : null;

		operand = f.operand;

		var_list = f.var_list != null ? new Vector(f.var_list) : null;

		predicate = f.predicate != null ? new Predicate(f.predicate) : null;

		// since Strings are static and will not be changed anyways, we
		// can just as well share the objects
		var_a = f.var_a;
		var_b = f.var_b;
		var_c = f.var_c;
		var_d = f.var_d;

		value = f.value != null ? new Kleene(f.value) : null;

	}
	
	public Object clone() {
		return new Formula(this);
	}

	public void replace(String a, String b) {
	}
	
	/* compare this formula with the given one. This does not do a full
		logical equality test but rather a literal comparison.
		Example: a -> b and (!a) | b would not be considered equal.
		However, some logical equalities are paid respect to:
			!E(v) x(v) <=> A(v) !x(v)
	*/
	public boolean equals(Formula f) {
		if (operand != f.operand)
			return isNegatedEqualQuantifier(f);
		if ((formula_a != null && !formula_a.equals(f.formula_a)) ||
			(formula_b != null && !formula_b.equals(f.formula_b)) ||
			(formula_c != null && !formula_c.equals(f.formula_c)))
			return false;
	
		if (var_list != null && f.var_list != null) {
			if (var_list.size() != f.var_list.size())
				return false;
			for (int i = 0; i < var_list.size(); i++)
				if (!var_list.elementAt(i).equals(f.var_list.elementAt(i)))
					return false;
		}
		if (predicate != null && !predicate.equals(f.predicate))
			return false;
		if ((var_a != null && !var_a.equals(f.var_a)) ||
			(var_b != null && !var_b.equals(f.var_b)) ||
			(var_c != null && !var_c.equals(f.var_c)) ||
			(var_d != null && !var_d.equals(f.var_d)))
			return false;
		if (value != null && !value.equals(f.value))
			return false;
		return true;
	}

	public boolean isQuantified() {
		return operand == FORALL || operand == EXISTS;
	}

	public boolean isNegatedEqualQuantifier(Formula f) {
		Formula negated = null, quantified = null;
		if (operand == NOT && f.isQuantified()) {
			negated = this;
			quantified = f;
		}
		else if (f.operand == NOT && isQuantified()) {
			negated = f;
			quantified = this;
		}
		if (negated == null || quantified == null)
			return false;
		return 
			negated.formula_a.isQuantified() &&
			negated.formula_a.operand != quantified.operand &&
			quantified.formula_a.operand == NOT;
	}

	/* creates a copy of this formula and logically inverts (negates) it.
		The inverted formula is returned, the actual formula remains
		unchanged.
	*/
	public Formula invert() {
		Formula f = new Formula();
		switch (operand) {
			case AMPERSAND:
				f.formula_a = formula_a.invert();
				f.formula_b = formula_b.invert();
				f.operand = OR;
				break;
			case OR:
				f.formula_a = formula_a.invert();
				f.formula_b = formula_b.invert();
				f.operand = AMPERSAND;
				break;
			case FOLLOWS:
				f.formula_a = (Formula)formula_a.clone();
				f.formula_b = formula_b.invert();
				f.operand = AMPERSAND;
				break;
			case FOLLOWS_BOTH:
				f.formula_a = (Formula)clone();
				f.operand = NOT;
				break;
			case NOT:
				f = (Formula)formula_a.clone();
				break;
			case QUESTION:
				f.formula_a = (Formula)clone();
				f.operand = NOT;
				break;
			case DOUBLE_EQUALS:
				f.var_a = var_a;	// String-sharing is safe
				f.var_b = var_b;
				f.operand = NOT_EQUAL;
				break;
			case NOT_EQUAL:
				f.var_a = var_a;
				f.var_b = var_b;
				f.operand = DOUBLE_EQUALS;
				break;
			case FORALL:
				f.formula_a = new Formula((Formula)formula_a.clone(), true);
				f.var_list = (Vector)var_list.clone();
				f.operand = EXISTS;
				break;
			case EXISTS:
				f.formula_a = new Formula((Formula)formula_a.clone(), true);
				f.var_list = (Vector)var_list.clone();
				f.operand = FORALL;
				break;
			// The next three can not be inverted XXX
			case PLUS:
			case MULTIPLY:
			case TC:
				f = (Formula)clone();
				break;
			case MAYBE:
				f.operand = MAYBE;
				switch (value.value) {
					case Kleene.TRUE:
						f.value = Kleene.KLEENE_FALSE;
						break;
					case Kleene.FALSE:
						f.value = Kleene.KLEENE_TRUE;
						break;
					case Kleene.MAYBE:
						f.value = Kleene.KLEENE_MAYBE;
						break;
				}
				break;
			default:
				// the only formula type without an operand set is a predicate
				// we don't want to invert anything here. Example:
				// this = a | b , inverted = !a & !b instead of !!a & !!b
				f = (Formula)clone();
		}
		return f;
	}

	public String toString() {
		switch (operand)
		{
			case AMPERSAND:
				return formula_a + " & " + formula_b;
			case OR:
				return formula_a + " | " + formula_b;
			case FOLLOWS:
				return formula_a + " -> " + formula_b;
			case FOLLOWS_BOTH:
				return formula_a + " <-> " + formula_b;
			case NOT:
				return "!" + formula_a;
			case QUESTION:
				return "(" +
					formula_a + " ? " +
					formula_b + " : " +
					formula_c + ")";
			case DOUBLE_EQUALS:
				return var_a + " == " + var_b;
			case NOT_EQUAL:
				return var_a + " != " + var_b;
			case FORALL:
				return "A(" + printVector(var_list) + ") " + formula_a;
			case EXISTS:
				return "E(" + printVector(var_list) + ") " + formula_a;
			case PLUS:
				return predicate + "+(" + var_a + ", " + var_b + ")";
			case MULTIPLY:
				return predicate + "*(" + var_a + ", " + var_b + ")";
			case TC:
				return "TC(" + var_a + ", " + var_b + ")(" + var_c + ", " + var_d +") " + formula_a;
			case MAYBE:
				return value.toString();
			default: // the only formula type without an operand set is a predicate
				return predicate + "(" + printVector(var_list, ',') + ")";
		}
	}
}
