package alexsa.gui;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Container;
import java.awt.GridLayout;
import javax.swing.JDialog;
import javax.swing.JInternalFrame;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.border.*;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

import alexsa.data.graphs.*;
import alexsa.data.tvp.*;
import alexsa.data.tvs.Structure;
import alexsa.data.animation.*;
import alexsa.event.*;
import alexsa.Alexsa;
import alexsa.Preferences;
import alexsa.WaitTimer;
import alexsa.gui.codeview.CodeView;
import alexsa.gui.menus.Commands;

public class AlexsaWindow
	extends JInternalFrame
	implements GraphListener, Runnable, Commands
{
	public final static boolean
		AUTO_SET_GRAPH = true,
		DONT_SET_GRAPH = false;

	private Program
		program = null;
	private Trace
		trace = null;
	private CodeView
		code_view = null;
	private GraphView
		graph_view = null;

	private Animation
		current_animation = null;
	
	private boolean
		animation_running = false;
		
	public AlexsaWindow(Program p, Trace t) {
		super(p.file.getName(), true, true, true, true);
		program = p;
		trace = t;
	
		setBackground(Color.white);
		
		Container content = getContentPane();
		content.setLayout(new GridLayout(1, 2));
		
		// left: the view for the control flow graph. The view is
		// wrapped in a JScrollPane, as to encompass larger graphs. The
		// JScrollPane is then wrapped within a standard JPanel, since
		// the border of a JScrollPane does not clear its background, which
		// would therefore remain gray.

		JPanel cfg_panel = new JPanel(new BorderLayout());
		cfg_panel.setBorder(new TitledBorder(new LineBorder(Color.black), "Program"));
		cfg_panel.setBackground(Color.white);
		
		code_view = new CodeView(program);
		code_view.addGraphListener(this);
		JScrollPane scroll_panel = new JScrollPane((Component)code_view);
		scroll_panel.setBorder(null);
		cfg_panel.add(scroll_panel, BorderLayout.CENTER);

		content.add(cfg_panel);

		// right: the current shape graph. Since the graph is automatically
		// scaled to its view's size, we don't need a JScrollPane
		JPanel panel = new JPanel(new BorderLayout());
		panel.setBorder(new TitledBorder(new LineBorder(Color.black), "Data"));
		panel.setBackground(getBackground());
		graph_view = new GraphView();
		panel.add(graph_view, BorderLayout.CENTER);
		content.add(panel);
	}

	/* while remaining at the same CFG node, show another shape graph for
	   the same CFG node.
	   Returns: true:  Another graph was found and was set
	            false: No other graph was found (to the right of this one).
	*/
	public boolean showPreviousGraph() {
		if (current_animation == null)
			return false;
		ShapeGraph sg = current_animation.getCurrentShapeGraph();
		CfgNode cn = code_view.getCurrentNode();
		ShapeGraph new_sg = cn.getPreviousShapeGraph(sg);
		if (new_sg != null) {
			syncAnimationWithGraph(new_sg);
			return true;
		}
		return false;
	}

	/* while remaining at the same CFG node, show another shape graph for
	   the same CFG node.
	   Returns: true:  Another graph was found and was set
	            false: No other graph was found (to the left of this one).
	*/
	public boolean showNextGraph() {
		if (current_animation == null)
			return false;
		ShapeGraph sg = current_animation.getCurrentShapeGraph();
		CfgNode cn = code_view.getCurrentNode();
		ShapeGraph new_sg = cn.getNextShapeGraph(sg);
		if (new_sg != null) {
			syncAnimationWithGraph(new_sg);
			return true;
		}
		return false;
	}

	public void set(CfgNode cn, ShapeGraph sg, boolean auto_set_graph, boolean jump) {
		if (current_animation != null && jump == false)
			graph_view.inbetweenToNewGraph(cn, sg);
		else
			graph_view.setGraph(cn, sg);
		activateNode(cn, auto_set_graph);
	}

	public boolean previous() {
		if (current_animation == null)
			return false;
		// we take our data from the current step, instead of going
		// back and getting the target of the step. This way, we
		// don't increase any pass-counters
		Step s = current_animation.getPreviousStep();
		if (s == null)
			return false;
		set(program.getCfgNode(s.shape_graph), s.shape_graph, DONT_SET_GRAPH, s.jump);
		return true;
	}
	
	public boolean next() {
		if (current_animation == null)
			return false;
		Step s = current_animation.getNextStep();
		if (s == null)
			return false;
		set(program.getCfgNode(s.shape_graph), s.shape_graph, DONT_SET_GRAPH, s.jump);
		return true;
	}

	public boolean startNewAnimation() {
		ShapeGraph[] gs = trace.initial_shapes;
		if (gs == null) {
			JOptionPane.showMessageDialog(this,
				"No initial shapes could be found, an animation is not possible.",
				"Not enough data", JOptionPane.ERROR_MESSAGE, null);
			return false;
		}
		int which = JOptionPane.CLOSED_OPTION;
		if (gs.length > 1) {
			StringBuffer b = new StringBuffer();
			String[] button_labels = new String[gs.length];
			for (int i = 0; i < gs.length; i++) {
				Structure st = (Structure)program.tvs.elementAt(i);
				st.analyze(program);
				b.append(st.getLongAnalysis());
				b.append('\n');
				if (st.description != null)
					button_labels[i] = gs[i].identifier + ": " + st.description;
				else
					button_labels[i] = "" + gs[i].identifier;
			}

			which = showOptionDialog(this, "Please select the shape you which to start with:\n\n" + b.toString(),
				"Multiple initial shapes", JOptionPane.DEFAULT_OPTION,
				JOptionPane.QUESTION_MESSAGE, 
				button_labels, button_labels[0]);
		} else
			which = 0;
		if (which != JOptionPane.CLOSED_OPTION) {
			current_animation = trace.animations[which];
			CfgNode cn = program.getCfgNode(gs[which]);
			activateNode(cn, DONT_SET_GRAPH);
			graph_view.setGraph(cn, gs[which]);
			return true;
		} else
			return false;
	}

	public void showAnimation() {
		if (startNewAnimation()) {
			animation_running = true;
			Thread t = new Thread(this);
			t.start();
		}
	}
	
	public void run() {
		do {
			wait(Preferences.speed);
		} while (animation_running && next());
		animation_running = false;
	}

	public void pauseContinueAnimation() {
		if (animation_running) {
			animation_running = false;
		} else {
			if (current_animation != null) {
				animation_running = true;
				Thread t = new Thread(this);
				t.start();
			} else
				showAnimation();
		}
	}

	public void wait(int ms) {
		long end = System.currentTimeMillis() + ms;
		while (System.currentTimeMillis() < end);
	}

	protected void activateNode(CfgNode cn, boolean auto_set_graph) {
		if (cn == code_view.getCurrentNode())
			return;
		code_view.activate(cn);

		updateMenus();
	}
	
	protected void syncAnimationWithNode(CfgNode n) {
		if (current_animation != null) {
			int i = current_animation.getFirstStepForCfgNode(n);
			if (i >= 0)
				current_animation.setAnimationTo(i);
			else
				current_animation = null;
		}
		for (int i = 0; i < trace.animations.length && current_animation == null; i++) {
			int j = trace.animations[i].getFirstStepForCfgNode(n);
			if (j >= 0) {
				current_animation = trace.animations[i];
				current_animation.setAnimationTo(j);
			}
		}
		if (current_animation != null)
			graph_view.setGraph(n, current_animation.getCurrentStep().shape_graph);
	}
	
	protected void syncAnimationWithGraph(ShapeGraph sg) {
		if (current_animation != null) {
			int i = current_animation.getFirstStepForShapeGraph(sg);
			if (i >= 0)
				current_animation.setAnimationTo(i);
			else
				current_animation = null;
		}
		for (int i = 0; i < trace.animations.length && current_animation == null; i++) {
			int j = trace.animations[i].getFirstStepForShapeGraph(sg);
			if (j >= 0) {
				current_animation = trace.animations[i];
				current_animation.setAnimationTo(j);
			}
		}
		if (current_animation != null)
			graph_view.setGraph(program.getCfgNode(sg), sg);
	}
	
	protected void showAction(alexsa.data.tvp.Action a) {
		if (a != null)
			JOptionPane.showMessageDialog(this,
				"TVP-Definition for the action \"" + a.identifier + "\":\n\n" + a.toString("    "),
				"Action " + a.identifier, JOptionPane.PLAIN_MESSAGE, null);
	}

	public void graphAction(GraphEvent ge) {
		if (ge.object instanceof CfgNode) {
			CfgNode cn = (CfgNode)ge.object;
			activateNode(cn, AUTO_SET_GRAPH);
			syncAnimationWithNode(cn);
		}
		else if (ge.object instanceof CfgEdge)
			showAction(program.getAction(((CfgEdge)ge.object).identifier));
	}

	private void updateMenus() {
		Component c = getParent();
		while (c != null && ! (c instanceof Alexsa))
			c = c.getParent();
		if (c != null)
			((Alexsa)c).updateMenus();
	}
	
	/* The Alexsa-Window will be asked by the application whether it can
		handle a certain menu-command. If that is the case, the AlexsaWindow
		will afterwards be insturcted to act on that command while the
		application will have no need to do further work on that command.
	*/
	public boolean canHandleCommand(int cmd) {
		int[] can_handle = {
			cmd_ViewSets, cmd_ViewCorePredicates,
			cmd_ViewInstrumentationPredicates, cmd_ViewConsistencyRules,
			cmd_ViewActions,
			cmd_up, cmd_down, cmd_forward, cmd_backward,
			cmd_Run, cmd_PauseContinueAnimation, cmd_DebugFunction,
			cmd_ExportCodeView, cmd_ExportTraversion};
		for (int i = 0; i < can_handle.length; i++)
			if (cmd == can_handle[i])
				return true;
		return false;
	}

	public void handleCommand(int cmd) {
		switch (cmd) {		
			case cmd_ViewSets:						showSets();						break;
			case cmd_ViewCorePredicates:			showCorePredicates();			break;
			case cmd_ViewInstrumentationPredicates:	showInstrumentationPredicates();break;
			case cmd_ViewConsistencyRules:			showConsistencyRules();			break;
			case cmd_ViewActions:					showActions();					break;
			case cmd_up:							previous();						break;
			case cmd_down:							next();							break;
			case cmd_forward:						showNextGraph();				break;
			case cmd_backward:						showPreviousGraph();			break;
			case cmd_Run:							showAnimation();				break;
			case cmd_PauseContinueAnimation:		pauseContinueAnimation();		break;
			case cmd_DebugFunction:					debugFunction();				break;
			case cmd_ExportCodeView:				exportCodeView();				break;
			case cmd_ExportTraversion:				exportAnimations();				break;
		}
	}
	
	public void exportCodeView() {
		javax.swing.JFileChooser chooser = new javax.swing.JFileChooser();
		chooser.setDialogTitle("Export code view");

		if (chooser.showSaveDialog(this) != javax.swing.JFileChooser.APPROVE_OPTION)
	    	return;
		
		java.io.File f = chooser.getSelectedFile();

		try {
			java.io.BufferedWriter bw = new java.io.BufferedWriter(new java.io.FileWriter(f));
			Vector v = code_view.getDisplayLines();
			for (int i = 0; i < v.size(); i++) {
				bw.write(v.elementAt(i).toString());
				bw.write('\n');
			}
			bw.close();
		} catch (java.io.IOException e) {
			System.err.println(e);
			e.printStackTrace();
		}
	}
	
	public void exportAnimations() {
		javax.swing.JFileChooser chooser = new javax.swing.JFileChooser();
		chooser.setDialogTitle("Export visual execution");

		if (chooser.showSaveDialog(this) != javax.swing.JFileChooser.APPROVE_OPTION)
	    	return;
		
		java.io.File f = chooser.getSelectedFile();

		try {
			java.io.BufferedWriter bw = new java.io.BufferedWriter(new java.io.FileWriter(f));
			trace.exportAnimations(bw);
			bw.close();
		} catch (java.io.IOException e) {
			System.err.println(e);
			e.printStackTrace();
		}
	}
	
	public void debugFunction() {
	}
	
	public static int showOptionDialog(
			Component parentComponent, Object message,
			String title, int optionType,
			int messageType,
			Object[] options, Object initialValue)
	{
		if (message instanceof String)
			message = (String) message + "\n ";

		final JOptionPane
			pane = new JOptionPane(message, messageType,
				optionType, null,
				options, initialValue);

		pane.setInitialValue(initialValue);

		final JDialog dialog = pane.createDialog(parentComponent, title);

		pane.selectInitialValue();
		dialog.setResizable(false);
		dialog.show();
		Object selectedValue = pane.getValue();
		dialog.dispose();

		if (selectedValue == null)
			return JOptionPane.CLOSED_OPTION;

		if (options == null) {
			if (selectedValue instanceof Integer)
				return ((Integer)selectedValue).intValue();
			return JOptionPane.CLOSED_OPTION;
		}
		for (int i = 0; i < options.length; i++) {
			if (options[i].equals(selectedValue))
				return i;
		}
		return JOptionPane.CLOSED_OPTION;
	}

	public void showSets() {
		if (program == null)
			return;

		StringBuffer b = new StringBuffer();
		Hashtable sets = program.getSets();
		if (sets == null)
			b.append("There are no sets declared in this .tvp-file.");
		else {
			if (sets.size() == 1)
				b.append("There is one set declared:\n");
			else {
				b.append("There are " + sets.size() +
					" sets declared:\n");
			}
			Enumeration e = sets.keys();
			while (e.hasMoreElements()) {
				String key = (String)e.nextElement();
				b.append("  ");
				b.append(key);
				b.append(" = ");
				b.append(sets.get(key));
				b.append('\n');
			}
		}
		JOptionPane.showMessageDialog(this, b.toString(),
			"Sets", JOptionPane.PLAIN_MESSAGE, null);
	}

	public void showCorePredicates() {
		if (program == null)
			return;

		StringBuffer b = new StringBuffer();
		Hashtable core_predicates = program.getCorePredicates();
		if (core_predicates == null)
			b.append("There are no core predicates declared in this .tvp-file.");
		else {
			if (core_predicates.size() == 1)
				b.append("There is one core predicate declared:\n");
			else {
				b.append("There are " + core_predicates.size() +
					" core predicates declared:\n");
			}
			Enumeration e = core_predicates.keys();
			while (e.hasMoreElements()) {
				b.append("  ");
				b.append(e.nextElement());
				b.append('\n');
			}
		}
		JOptionPane.showMessageDialog(this, b.toString(),
			"Core predicates", JOptionPane.PLAIN_MESSAGE, null);
	}

	public void showInstrumentationPredicates() {
		if (program == null)
			return;

		StringBuffer b = new StringBuffer();
		Hashtable instrumentation_predicates = program.getInstrumentationPredicates();
		if (instrumentation_predicates == null)
		{
			b.append("There are no instrumentation predicates declared in this .tvp-file.");
		} else {
			if (instrumentation_predicates.size() == 1)
				b.append("There is one instrumentation predicate declared:\n");
			else {
				b.append("There are " + instrumentation_predicates.size() +
					" instrumentation predicates declared:\n");
			}
			Enumeration e = instrumentation_predicates.elements();
			while (e.hasMoreElements()) {
				b.append("  ");
				b.append(e.nextElement());
				b.append('\n');
			}
		}
		JOptionPane.showMessageDialog(this, b.toString(),
			"Instrumentation predicates", JOptionPane.PLAIN_MESSAGE, null);
	}

	public void showConsistencyRules() {
		if (program == null)
			return;

		StringBuffer b = new StringBuffer();
		Hashtable consistency_rules = program.getConsistencyRules();
		if (consistency_rules == null)
		{
			b.append("There are no consistency rules declared in this .tvp-file.");
		} else {
			if (consistency_rules.size() == 1)
				b.append("There is one consistency rule explicitly declared:\n");
			else {
				b.append("There are " + consistency_rules.size() +
					" consistency rules explicitly declared:\n");
			}
			Enumeration e = consistency_rules.elements();
			while (e.hasMoreElements()) {
				b.append("  ");
				b.append(e.nextElement());
				b.append('\n');
			}
		}
		JOptionPane.showMessageDialog(this, b.toString(),
			"Consistency rules", JOptionPane.PLAIN_MESSAGE, null);
	}
	
	public void showActions() {
		if (program == null)
			return;
		Object[] actions = program.getAllActions();
		String[] names = new String[actions.length];
		for (int i = actions.length - 1; i >= 0; i--)
			names[i] = ((Action)actions[i]).identifier;
		String action_name = (String)JOptionPane.showInputDialog(this, 
			"Select an action", "Input",
			JOptionPane.INFORMATION_MESSAGE, null,
			names, names[0]);
		if (action_name != null) {
			int i = 0;
			Action action = null;
			while (i < actions.length && action == null) {
				if (names[i] == action_name)
					action = (Action)actions[i];
				else
					i++;
			}
			String action_text = action.toString();
			StringBuffer b = new StringBuffer();
			i = 0;
			char c;
			while (i < action_text.length()) {
				c = action_text.charAt(i++);
				if (c == '\t')
					b.append("    ");
				else
					b.append(c);
			}
			JOptionPane.showMessageDialog(this,
				b.toString(),
				"Action [" + action_name + "]",
				JOptionPane.PLAIN_MESSAGE, null);
		}
	}
}
