package alexsa.gui;

import javax.swing.*;
import javax.swing.border.*;
import java.awt.*;
import java.awt.geom.*;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.util.Vector;

import alexsa.data.graphs.*;
import alexsa.data.morphing.*;
import alexsa.data.animation.Transition;
import alexsa.data.tvp.CfgNode;
import alexsa.data.Edge;
import alexsa.data.Node;
import alexsa.data.Tools;
import alexsa.event.GraphEvent;
import alexsa.event.GraphListener;
import alexsa.WaitTimer;
import alexsa.Preferences;

public class GraphView
	extends JComponent implements Runnable
{

	protected final static String
		EMPTY_GRAPH_MESSAGE = "(Empty graph)";

	protected static int
		FONT_SIZE = 12;

	protected ShapeGraph
		graph = null;
	protected AffineTransform
		scaling = null;
	protected boolean
		determine_scaling = true;
	protected FontMetrics
		font_metrics = null;
	private Image
		offscreen_image = null;
	private Graphics2D
		offscreen_graphics = null;
	
		
	protected BasicStroke
		// The Basic Stroke always remains the same, we can therefore
		// define it right here
		SolidStroke = new BasicStroke(),
		// The dotted stroke depends on the scaling factor and is thus
		// recalculated every time a new graph is set and/or the
		// component is resized
		DottedStroke = null;

	public GraphView() {
		super();
		setBackground(Color.white);
		addComponentListener(new ComponentAdapter() {
			public void componentResized(ComponentEvent e) {
				determine_scaling = true;
				offscreen_image = null;
				offscreen_graphics = null;
				repaint();
			}
		});
		setDoubleBuffered(false);
	}
	
	private void setFont(Graphics g) {
		int font_size = FONT_SIZE;
		if (scaling != null)
			font_size = (int)((double)FONT_SIZE / scaling.getScaleX());
		Font font = new Font("SansSerif", Font.PLAIN, font_size);
		if (g != null)
			font_metrics = g.getFontMetrics(font);
		else
			font_metrics = null;
		setFont(font);
		if (offscreen_graphics != null)
			offscreen_graphics.setFont(font);
	}

	private AffineTransform determineScaling(Graphics2D g, ShapeGraph graph) {
		determine_scaling = false;
		AffineTransform scaling = null;
		if (graph == null || graph.width == 0) {
			setFont(g);
			return null;
		}
		Dimension size = getSize();
		double
			X = (double)size.width / (double)graph.width,
			Y = (double)size.height / (double)graph.height,
			Factor = Math.min(X, Y);

		// we scale by the appropriate factor and then we center the graph.
		// To do this, we use the
		scaling = new AffineTransform(
			Factor, 0,
			0, Factor,
			(size.width - graph.width * Factor) / 2,
			(size.height - graph.height * Factor) / 2);

		float[] DotPattern = {
			(float)(5 / Factor),
			(float)(5 / Factor)};
		if (DotPattern[0] == 0) {
			System.out.println("Hm? " + Factor + " [" + graph.width + ", " +
				graph.height + "]");
		}
		DottedStroke = new BasicStroke(SolidStroke.getLineWidth(), SolidStroke.getEndCap(),
			   SolidStroke.getLineJoin(), SolidStroke.getMiterLimit(),
			   DotPattern, 0);
		return scaling;
	}

	public void setGraph(CfgNode cn, int index) {
		ShapeGraph g = (ShapeGraph)cn.shape_graphs.elementAt(index);
		setTitle(cn, g, index);
		setGraph(g);
	}

	public void setGraph(CfgNode cn, ShapeGraph g) {
		setTitle(cn, g, cn.shape_graphs.indexOf(g));
		setGraph(g);
	}

	private void setTitle(CfgNode cn, ShapeGraph g, int index) {
		String title = null;
		if (g.identifier >= 0)
			title = "Data for state \"" + cn.name +
				"\", graph #" + g.identifier + " [" +
				(index + 1) + " / " + cn.shape_graphs.size() + "]";
		else
			title = "Data for state \"" + cn.name +
				"\" [" + (index + 1) + " / " + cn.shape_graphs.size() + "]";

		Border border = getBorder();
		if (border == null) {
			Container c = getParent();
			if (c != null && c instanceof JPanel)
				border = ((JPanel)c).getBorder();
		}
		if (border != null && border instanceof TitledBorder) {
			((TitledBorder)border).setTitle(title);
			getParent().repaint();
		}
	}

	private void setGraph(ShapeGraph g) {
		graph = g;
		determine_scaling = true;
		repaint(0);
	}

	protected void drawNode(GraphNode gn, Color c) {
		if (gn == null)
			return;
		if (c == null) {
			if (gn.active)
				c = Color.red;
			else
				c = Color.black;
		}
		offscreen_graphics.setColor(c);
		int shape = gn.getShape();
		if (gn.attributes != null) {
			if (gn.attributes.isRound())
				shape = GraphSymbols.ELLIPSE;
			else if (gn.attributes.isRectangular())
				shape = GraphSymbols.BOX;
		}
			
		if (gn.isDotted())
			offscreen_graphics.setStroke(DottedStroke);
		switch (shape) {	
			case GraphSymbols.ELLIPSE:
				offscreen_graphics.drawOval(gn.x - gn.width / 2, gn.y - gn.height / 2, gn.width, gn.height);
				break;
			case GraphSymbols.BOX:
				offscreen_graphics.drawRect(gn.x - gn.width / 2, gn.y - gn.height / 2, gn.width, gn.height);
				break;
		}
		if (gn.isDotted())
			offscreen_graphics.setStroke(SolidStroke);
		
		if (gn.label.indexOf("\\n") <= 0)
			try {
				offscreen_graphics.drawString(gn.label,
					gn.x - font_metrics.stringWidth(gn.label) / 2,
					gn.y + font_metrics.getAscent() / 2);
			} catch (IllegalArgumentException exc) {
				exc.printStackTrace();
			}
		else {
			int labelx, labely;
			String[] Lines = Tools.split(gn.label, "\\n");

			// first: determine center of node in screen coordinates
			labely = gn.y
				// then consider number of lines
				- (Lines.length * font_metrics.getHeight()) / 2
				// .. and finally the offset of one line
				+ font_metrics.getAscent();
			try {
				for (int i = 0; i < Lines.length; i++) {
					labelx = gn.x - font_metrics.stringWidth(Lines[i]) / 2;
					offscreen_graphics.drawString(Lines[i], labelx, labely + i * font_metrics.getHeight());
				}
			} catch (IllegalArgumentException exc) {
				exc.printStackTrace();
			}
		}
	}

	protected void drawEdge(GraphEdge e, Color c) {
		if (c != null)
			offscreen_graphics.setColor(c);
		else {
			if (e.attributes != null && e.attributes.color == GraphSymbols.RED)
				offscreen_graphics.setColor(Color.red);
			else
				offscreen_graphics.setColor(Color.black);
		}

		if (e.isDotted())
			offscreen_graphics.setStroke(DottedStroke);
			
		CubicCurve2D curve = new CubicCurve2D.Double();
		for (int j = 0; j < (e.n - 1) / 3; j++) {
			Point
				p1 = (Point)e.points.elementAt(j * 3),
				p2 = (Point)e.points.elementAt(j * 3 + 1),
				p3 = (Point)e.points.elementAt(j * 3 + 2),
				p4 = (Point)e.points.elementAt(j * 3 + 3);
			curve.setCurve(p1.x, p1.y, p2.x, p2.y, p3.x, p3.y, p4.x, p4.y);
			offscreen_graphics.draw(curve);
		}
		if (e.isDotted())
			offscreen_graphics.setStroke(SolidStroke);
		String label = e.getLabel();
		if (label != null)
			try {
				offscreen_graphics.drawString(label,
					e.label.location.x - font_metrics.stringWidth(label) / 2,
					e.label.location.y - font_metrics.getHeight() / 2 + font_metrics.getAscent());
			} catch (IllegalArgumentException exc) {
				exc.printStackTrace();
			}

		drawArrow(offscreen_graphics, (Point)e.points.elementAt(e.n - 1), (Point)e.points.elementAt(e.n), !e.isDotted());
	}
	
	// The graphs generated by dot do not contain appropriate arrows
	// connecting the edges with the target nodes. Therefore we draw our
	// own arrows.
	protected void drawArrow(Graphics2D g2d, Point arrow_root, Point arrow_point, boolean filled) {
		double arrow_width = .40,
			angle = Math.atan(
				(float)(arrow_root.y - arrow_point.y) /
				(float)(arrow_root.x - arrow_point.x)),

			x = arrow_point.x,
			y = arrow_point.y,
			// the length of the arrow
			length = arrow_root.distance(x, y),

			// vertical and horizontal distances of the "outer" points
			// of the arrow
			dx = length * arrow_width * Math.cos(angle - Math.PI / 2) / 2,
			dy = length * arrow_width * Math.sin(angle - Math.PI / 2) / 2;

		GeneralPath arrow = new GeneralPath();
		arrow.moveTo((float)x, (float)y);
		arrow.lineTo((float)(arrow_root.x - dx), (float)(arrow_root.y - dy));
		arrow.lineTo((float)(arrow_root.x + dx), (float)(arrow_root.y + dy));
		arrow.lineTo((float)x, (float)y);
		if (filled)
			g2d.fill(arrow);
		else
			g2d.draw(arrow);
	}

	protected Rectangle getTransformedSize() {
		Dimension size = getSize();
		if (scaling == null)
			return new Rectangle(0, 0, size.width, size.height);

		double
			factor = scaling.getScaleX(),
			dst[] = {
				0.0 - scaling.getTranslateX() / factor,
				0.0 - scaling.getTranslateY() / factor,
				(float)size.width / factor,
				(float)size.height / factor
			};
		return new Rectangle((int)dst[0], (int)dst[1], (int)dst[2], (int)dst[3]);
	}
	
	
	public void paintImage() {
		if (determine_scaling) {
			scaling = determineScaling(offscreen_graphics, graph);
			// the scaling is null for empty graphs
			if (scaling != null)
				offscreen_graphics.setTransform(scaling);
			else
				offscreen_graphics.setTransform(new AffineTransform());
			setFont(offscreen_graphics);
		}

		offscreen_graphics.setColor(getBackground());
		
		Rectangle full_area = getTransformedSize();
		offscreen_graphics.setClip(full_area.x, full_area.y, full_area.width, full_area.height);
		offscreen_graphics.setClip(full_area.x, full_area.y, full_area.width + Math.abs(full_area.x), full_area.height + Math.abs(full_area.y));
		offscreen_graphics.fillRect(full_area.x, full_area.y, full_area.width, full_area.height);
		
		offscreen_graphics.setColor(Color.black);

		// paint all nodes
		for (int i = graph.getNumberOfNodes() - 1; i >= 0; i --)
			drawNode(graph.getNode(i), null);

		// paint all edges
		for (int i = graph.getNumberOfEdges() - 1; i >= 0; i --)
			drawEdge(graph.getEdge(i), null);
	
		// if we haven't drawn anythig else, we print out a little
		// "empty graph"-message
		if (graph.getNumberOfNodes() == 0 && graph.getNumberOfEdges() == 0)
			offscreen_graphics.drawString(EMPTY_GRAPH_MESSAGE,
				(full_area.width - font_metrics.stringWidth(EMPTY_GRAPH_MESSAGE)) / 2,
				(full_area.height - font_metrics.getHeight()) / 2);
	}
	
	/* the next two methods handle the multithreading
	  in swing drawing. Beating multithreading is important because
	  Swing is not thread safe. Without this serious redraw problems
	  would arise
	*/
	private Graphics
		paint_graphics = null;

	public void run() {
		paint_graphics.drawImage(offscreen_image, 0, 0, this);
	}
	
	private void paintItNow(Graphics g) {
		if (SwingUtilities.isEventDispatchThread()) {
			g.drawImage(offscreen_image, 0, 0, this);
		} else {
			paint_graphics = g;
			try {
				javax.swing.SwingUtilities.invokeAndWait(this);
			} catch (java.lang.reflect.InvocationTargetException e) {
			} catch (InterruptedException e) {
			}
		}
	}

	public void paintComponent(Graphics g) {
		if (graph == null)
			return;
		Dimension size = getSize();
		if (offscreen_image == null) {
			offscreen_image = createImage(getSize().width, getSize().height);
			offscreen_graphics = (Graphics2D)offscreen_image.getGraphics();
		}
		
		paintImage();
		paintItNow(g);
	}

	public void inbetweenToNewGraph(CfgNode cn, ShapeGraph new_graph) {
		Transition original_t = null;
		Edge e = null;
		boolean stepping_back = false;

		// set the title "Data for "n_x", graph #a [b/c]"
		setTitle(cn, new_graph, cn.shape_graphs.indexOf(new_graph));

		if (graph != null) {
			for (int i = 0; original_t == null && i < graph.outDegree(); i++)
				if ((e = graph.getOutEdge(i)).target == new_graph)
					original_t = (Transition)e;
			for (int i = 0; original_t == null && i < graph.inDegree(); i++)
				if ((e = graph.getInEdge(i)).source == new_graph) {
					original_t = (Transition)e;
					stepping_back = true;
				}
		}
		// if there is no transition connecting the current shape graph
		// with the new shape graph, then we cannot perform an inbetweening.
		// we just set the new graph abruptly and return
		// we also do this in case of a back-edge, as we cannot safely
		// determine a well-defined inbetweening
		if (original_t == null || original_t.type == Edge.BACK_EDGE) {
			setGraph(new_graph);
			return;
		}

		Transition t = new Transition(
			(ShapeGraph)((ShapeGraph)original_t.source).clone(), 
			(ShapeGraph)((ShapeGraph)original_t.target).clone());
		t.type = original_t.type;
		Morphing morphing = new Morphing(t);

		if (morphing.is_broken) {
//			System.out.println("broken transition");
			setGraph(new_graph);
			return;
		}

		Rectangle full_area = getTransformedSize();
		offscreen_graphics.setClip(full_area.x, full_area.y, full_area.width, full_area.height);

		Graphics2D g = (Graphics2D)getGraphics();
		AffineTransform
			new_scaling = determineScaling(g, new_graph);
		Dimension size = getSize();

		if (scaling != null && new_scaling != null &&
			scaling.getScaleX() > new_scaling.getScaleX())
		{
			zoom(g, scaling, new_scaling, size);
		}
	
		// transform....

		removeNodes(g, morphing.removed);
		mergeNodes(g, morphing.merges);
		moveNodes(g, morphing.moves);
		splitNodes(g, morphing.splits);
		createNodes(g, morphing.created);
		
		renameNodes(g, morphing.label_changes);

/* original order:
		removeNodes(g, morphing.removed);
		moveNodes(g, morphing.moves);
		createNodes(g, morphing.created);
		mergeNodes(g, morphing.merges);
		splitNodes(g, morphing.splits);
		renameNodes(g, morphing.label_changes);
*/
		if (scaling != null && new_scaling != null &&
			scaling.getScaleX() < new_scaling.getScaleX())
		{
			graph = new_graph;
			zoom(g, scaling, new_scaling, size);
		}

		WaitTimer.waitFor(Preferences.FINAL_DELAY);

		g.dispose();

		setGraph(new_graph);
	}
	
	protected void zoom(
		Graphics2D g,
		AffineTransform scaling,
		AffineTransform new_scaling,
		Dimension size)
	{
		double old_factor = scaling.getScaleX(),
			new_factor = new_scaling.getScaleX(),
			old_translate_x = scaling.getTranslateX(),
			old_translate_y = scaling.getTranslateY(),
			new_translate_x = new_scaling.getTranslateX(),
			new_translate_y = new_scaling.getTranslateY();
		for (int i = 1; i <= Preferences.ZOOM_STEPS; i++) {
			WaitTimer.mark();
			double factor = old_factor + (new_factor - old_factor) * i / Preferences.ZOOM_STEPS;
			scaling.setTransform(
				factor, 0, 0, factor,
				old_translate_x + (new_translate_x - old_translate_x) * i / Preferences.ZOOM_STEPS,
				old_translate_y + (new_translate_y - old_translate_y) * i / Preferences.ZOOM_STEPS);
			offscreen_graphics.setTransform(scaling);
			setFont(offscreen_graphics);
			paintImage();
			paintItNow(g);

			WaitTimer.wait(Preferences.ZOOM_DELAY);
		}
		scaling = new_scaling;
		setFont(offscreen_graphics);
		offscreen_graphics.setTransform(scaling);
		paintImage();
		paintItNow(g);
	}

	protected void removeNodes(Graphics g, Vector nodes) {
		if (nodes == null)
			return;
		for (int i = Preferences.REMOVE_STEPS - 1; i >= 0; i--) {
			WaitTimer.mark();
			int c = 255 - (255 * i) / Preferences.REMOVE_STEPS;
			Color color = new Color(c, c, c);
			for (int j = nodes.size() - 1; j >= 0; j--) {
				GraphNode gn = (GraphNode)nodes.elementAt(j);
				drawNode(gn, color);
				drawAllEdges(gn, color);
			}
			paintItNow(g);
			WaitTimer.wait(Preferences.REMOVE_DELAY);
		}
		for (int j = nodes.size() - 1; j >= 0; j--) {
			GraphNode gn = (GraphNode)nodes.elementAt(j);
			for (int k = gn.inDegree() - 1; k >= 0; k--) {
				Edge e = gn.getInEdge(k);
				Node source = e.source;
				source.removeOutEdge(e);
			}
			for (int k = gn.outDegree() - 1; k >= 0; k--) {
				Edge e = gn.getOutEdge(k);
				Node target = e.target;
				target.removeInEdge(e);
			}
		}
	}
	
	protected void moveNodes(Graphics g, Vector nodes) {
		if (nodes == null)
			return;
		// first, we erase all existing edges around the node, as they
		// may be slightly different from the later drawn "simple" edges
		for (int j = nodes.size() - 1; j >= 0; j--) {
			Move move = (Move)nodes.elementAt(j);
			drawAllEdges(move.source, getBackground());
		}
		paintItNow(g);

		for (int i = Preferences.MOVE_STEPS; i > 0; i--) {
			WaitTimer.mark();
			for (int j = nodes.size() - 1; j >= 0; j--) {
				Move move = (Move)nodes.elementAt(j);
				int dx = (move.target.x - move.source.x) / i,
					dy = (move.target.y - move.source.y) / i;
				moveNode(move.source, dx, dy);
			}
			paintItNow(g);
			WaitTimer.wait(Preferences.MOVE_DELAY);
		}
		paintItNow(g);
	}
	
	protected void moveNode(GraphNode gn, int dx, int dy) {
		drawNode(gn, getBackground());

		for (int i = gn.inDegree() - 1; i >= 0; i--) {
			GraphEdge ge = (GraphEdge)gn.getInEdge(i);
			if (ge.source != ge.target)
				drawSimpleEdge((GraphNode)ge.source, (GraphNode)ge.target, ge, getBackground());
			else {
				drawEdge(ge, getBackground());
				moveEdge(ge, gn, dx, dy);
				drawEdge(ge, null);
			}
		}
		for (int i = gn.outDegree() - 1; i >= 0; i--) {
			GraphEdge ge = (GraphEdge)gn.getOutEdge(i);
			if (ge.source != ge.target)
				drawSimpleEdge((GraphNode)ge.source, (GraphNode)ge.target, ge, getBackground());
		}
		gn.x += dx;
		gn.y += dy;
		drawNode(gn, null);

		for (int i = gn.inDegree() - 1; i >= 0; i--) {
			GraphEdge ge = (GraphEdge)gn.getInEdge(i);
			if (ge.source != ge.target)
				drawSimpleEdge((GraphNode)ge.source, (GraphNode)ge.target, ge, null);
		}
		for (int i = gn.outDegree() - 1; i >= 0; i--) {
			GraphEdge ge = (GraphEdge)gn.getOutEdge(i);
			if (ge.source != ge.target)
				drawSimpleEdge((GraphNode)ge.source, (GraphNode)ge.target, ge, null);
		}
	}
	
	protected void drawAllEdges(GraphNode gn, Color c) {
		for (int i = gn.inDegree() - 1; i >= 0; i--)
			drawEdge((GraphEdge)gn.getInEdge(i), c);
		for (int i = gn.outDegree() - 1; i >= 0; i--)
			drawEdge((GraphEdge)gn.getOutEdge(i), c);
	}
	
	protected void drawAllSimpleEdges(GraphNode gn, Color c) {
		for (int i = gn.inDegree() - 1; i >= 0; i--) {
			GraphEdge ge = (GraphEdge)gn.getInEdge(i);
			if (ge.source != ge.target)
				drawSimpleEdge((GraphNode)ge.source, (GraphNode)ge.target, ge, c);
			else
				drawEdge(ge, c);
		}
		for (int i = gn.outDegree() - 1; i >= 0; i--) {
			GraphEdge ge = (GraphEdge)gn.getOutEdge(i);
			if (ge.source != ge.target)
				drawSimpleEdge((GraphNode)ge.source, (GraphNode)ge.target, ge, c);
		}
	}
	
	protected int
		DEFAULT_ARROW_LENGTH = 140;
	
	protected void drawSimpleEdge(GraphNode a, GraphNode b, GraphEdge e, Color c) {
		if (c != null)
			offscreen_graphics.setColor(c);
		else {
			if (e.attributes != null && e.attributes.color == GraphSymbols.RED)
				offscreen_graphics.setColor(Color.red);
			else
				offscreen_graphics.setColor(Color.black);
		}

		if (e.isDotted())
			offscreen_graphics.setStroke(DottedStroke);
		
		Point p_a = a.getPointOnRimTowards(b, 0),
			p_middle = b.getPointOnRimTowards(a, DEFAULT_ARROW_LENGTH),
			p_b = b.getPointOnRimTowards(a, 0);
		offscreen_graphics.drawLine(p_a.x, p_a.y, p_middle.x, p_middle.y);

		if (e.isDotted())
			offscreen_graphics.setStroke(SolidStroke);

		drawArrow(offscreen_graphics, p_middle, p_b, !e.isDotted());
		
		if (e.label != null) {
			Point p_label = getLabelPosition(p_a, p_b);
			offscreen_graphics.drawString(e.label.label, p_label.x, p_label.y);
		}
	}
	
	protected void mergeNodes(Graphics g, Vector merges) {
		if (merges == null)
			return;
		// Step 1: init the megre
		for (int j = merges.size() - 1; j >= 0; j--) {
			Merge merge = (Merge)merges.elementAt(j);
			merge.init();
		}

		// Step 2: remove old edges
		for (int j = merges.size() - 1; j >= 0; j--) {
			Merge merge = (Merge)merges.elementAt(j);
			if (merge.old_edges.size() == 0)
				continue;
			for (int i = Preferences.REMOVE_STEPS - 1; i >= 0; i--) {
				WaitTimer.mark();
				int c = 255 - (255 * i) / Preferences.REMOVE_STEPS;
				Color color = new Color(c, c, c);
				for (int k = merge.old_edges.size() - 1; k >= 0; k--) {
					GraphEdge ge = (GraphEdge)merge.old_edges.elementAt(k);
					drawEdge(ge, color);
				}
				paintItNow(g);
				WaitTimer.wait(Preferences.REMOVE_DELAY);
			}
		}
	
		// Step 3: erase the old nodes
		for (int j = merges.size() - 1; j >= 0; j--) {
			Merge merge = (Merge)merges.elementAt(j);
			drawNode(merge.source, getBackground());
			drawAllEdges(merge.source, getBackground());
			drawAllSimpleEdges(merge.source, getBackground());
			drawNode(merge.source2, getBackground());
			drawAllEdges(merge.source2, getBackground());
			drawAllSimpleEdges(merge.source2, getBackground());
		}
		
		// Step 4: merge the nodes
		for (int i = Preferences.MERGE_STEPS; i > 0; i--) {
			WaitTimer.mark();
			for (int j = merges.size() - 1; j >= 0; j--) {
				Merge merge = (Merge)merges.elementAt(j);

				// draw all internal edges in white
				if (!merge.sources[0].overlapsWith(merge.sources[1]))
					for (int k = 0; k < merge.internal_edges.size(); k ++) {
						GraphEdge internal_edge = (GraphEdge)merge.internal_edges.elementAt(k);
						if (internal_edge.source == merge.source)
							drawSimpleEdge(merge.sources[0], merge.sources[1], internal_edge, getBackground());
						else if (internal_edge.source == merge.source2)
							drawSimpleEdge(merge.sources[1], merge.sources[0], internal_edge, getBackground());
					}

				// move the node and all connected edges
				for (int m = 0; m < 2; m++)
					moveNode(merge.sources[m],
						(merge.target.x - merge.sources[m].x) / i,
						(merge.target.y - merge.sources[m].y) / i);

				// now draw all new internal edges
				if (!merge.sources[0].overlapsWith(merge.sources[1]))
					for (int k = 0; k < merge.internal_edges.size(); k ++) {
						GraphEdge internal_edge = (GraphEdge)merge.internal_edges.elementAt(k);
						if (internal_edge.source == merge.source)
							drawSimpleEdge(merge.sources[0], merge.sources[1], internal_edge, null);
						else if (internal_edge.source == merge.source2)
							drawSimpleEdge(merge.sources[1], merge.sources[0], internal_edge, null);
					}
			}
			paintItNow(g);
			WaitTimer.wait(Preferences.MERGE_DELAY);
		}

		// Step 5: delete the two old nodes, draw new node
		for (int j = merges.size() - 1; j >= 0; j--) {
			Merge merge = (Merge)merges.elementAt(j);
			for (int i = 0; i < 2; i++) {
				drawNode(merge.sources[i], getBackground());
				drawAllEdges(merge.sources[i], getBackground());
				drawAllSimpleEdges(merge.sources[i], getBackground());
			}
			drawNode(merge.target, null);
			drawAllEdges(merge.target, null);
			if (merge.new_edges.size() > 0)
				for (int i = Preferences.CREATE_STEPS - 1; i >= 0; i--) {
					WaitTimer.mark();
					int c = (255 * i)/ Preferences.CREATE_STEPS;
					Color color = new Color(c, c, c);
					for (int k = merge.new_edges.size() - 1; k >= 0; k--) {
						GraphEdge ge = (GraphEdge)merge.new_edges.elementAt(k);
						drawEdge(ge, color);
					}
					paintItNow(g);
					WaitTimer.wait(Preferences.CREATE_DELAY);
				}
			paintItNow(g);
		}
	}

	protected void splitNodes(Graphics g, Vector splits) {
		if (splits == null)
			return;

		// Step 1: init the splits
		for (int j = splits.size() - 1; j >= 0; j--) {
			Split split = (Split)splits.elementAt(j);
			split.init();
		}

		// Step 2: remove old edges
		for (int j = splits.size() - 1; j >= 0; j--) {
			Split split = (Split)splits.elementAt(j);
			if (split.old_edges.size() == 0)
				continue;
			for (int i = Preferences.REMOVE_STEPS - 1; i >= 0; i--) {
				WaitTimer.mark();
				int c = 255 - (255 * i) / Preferences.REMOVE_STEPS;
				Color color = new Color(c, c, c);
				for (int k = split.old_edges.size() - 1; k >= 0; k--) {
					GraphEdge ge = (GraphEdge)split.old_edges.elementAt(k);
					drawEdge(ge, color);
				}
				paintItNow(g);
				WaitTimer.wait(Preferences.REMOVE_DELAY);
			}
		}
	
		// Step 3: erase the old node
		for (int j = splits.size() - 1; j >= 0; j--) {
			Split split = (Split)splits.elementAt(j);
			drawNode(split.source, getBackground());
			// used to be: draw only normal edges, but since we might have
			// simple edges from a previous move, we draw them, too.
			drawAllEdges(split.source, getBackground());
			drawAllSimpleEdges(split.source, getBackground());
		}
		
		// Step 4: split the node
		for (int i = Preferences.SPLIT_STEPS; i > 0; i--) {
			WaitTimer.mark();
			for (int j = splits.size() - 1; j >= 0; j--) {
				Split split = (Split)splits.elementAt(j);

				// draw all internal edges in white
				if (!split.sources[0].overlapsWith(split.sources[1]))
					for (int k = 0; k < split.internal_edges.size(); k ++) {
						GraphEdge internal_edge = (GraphEdge)split.internal_edges.elementAt(k);
						if (internal_edge.source == split.targets[0])
							drawSimpleEdge(split.sources[0], split.sources[1], internal_edge, getBackground());
						else if (internal_edge.source == split.targets[1])
							drawSimpleEdge(split.sources[1], split.sources[0], internal_edge, getBackground());
					}

				// move the node and all connected edges
				for (int m = 0; m < 2; m++)
					moveNode(split.sources[m],
						(split.targets[m].x - split.sources[m].x) / i,
						(split.targets[m].y - split.sources[m].y) / i);

				// now draw all new internal edges
				if (!split.sources[0].overlapsWith(split.sources[1]))
					for (int k = 0; k < split.internal_edges.size(); k ++) {
						GraphEdge internal_edge = (GraphEdge)split.internal_edges.elementAt(k);
						if (internal_edge.source == split.targets[0])
							drawSimpleEdge(split.sources[0], split.sources[1], internal_edge, null);
						else if (internal_edge.source == split.targets[1])
							drawSimpleEdge(split.sources[1], split.sources[0], internal_edge, null);
					}
			}
			paintItNow(g);
			WaitTimer.wait(Preferences.SPLIT_DELAY);
		}

		// step 5: create new edges
		for (int j = splits.size() - 1; j >= 0; j--) {
			Split split = (Split)splits.elementAt(j);
			if (split.new_edges.size() == 0)
				continue;
			for (int i = Preferences.CREATE_STEPS - 1; i >= 0; i--) {
				WaitTimer.mark();
				int c = (255 * i)/ Preferences.CREATE_STEPS;
				Color color = new Color(c, c, c);
				for (int k = split.new_edges.size() - 1; k >= 0; k--) {
					GraphEdge ge = (GraphEdge)split.new_edges.elementAt(k);
					drawEdge(ge, color);
				}
				paintItNow(g);
				WaitTimer.wait(Preferences.CREATE_DELAY);
			}
		}
	
	}

	protected void renameNodes(Graphics g, Vector nodes) {
		if (nodes == null)
			return;
		for (int i = nodes.size() - 1; i >= 0; i--) {
			Color color = getBackground();
			for (GraphNode gn = ((Rename)nodes.elementAt(i)).source;
				gn != null;
				gn = gn == ((Rename)nodes.elementAt(i)).source ? ((Rename)nodes.elementAt(i)).target : null)
			{
				drawNode(gn, color);
				for (int k = 0; k < gn.outDegree(); k++)
					drawEdge((GraphEdge)gn.getOutEdge(k), color);
				for (int k = 0; k < gn.inDegree(); k++)
					drawEdge((GraphEdge)gn.getInEdge(k), color);
				color = null;
			}
		}
	}
	
	protected void createNodes(Graphics g, Vector nodes) {
		if (nodes == null)
			return;
		for (int i = Preferences.CREATE_STEPS - 1; i >= 0; i--) {
			int c = (255 * i)/ Preferences.CREATE_STEPS;
			Color color = new Color(c, c, c);
			for (int j = nodes.size() - 1; j >= 0; j--) {
				WaitTimer.mark();
				GraphNode gn = (GraphNode)nodes.elementAt(j);
				drawNode(gn, color);
				for (int k = gn.outDegree() - 1; k >= 0; k--)
					drawEdge((GraphEdge)gn.getOutEdge(k), color);
			}
			paintItNow(g);
			WaitTimer.wait(Preferences.CREATE_DELAY);
		}
	}
	
	protected void moveAllEdges(GraphNode gn, int dx, int dy) {
		for (int i = 0; i < gn.outDegree(); i++) {
			GraphEdge ge = (GraphEdge)gn.getOutEdge(i);
			drawEdge(ge, getBackground());
			moveEdge(ge, gn, dx, dy);
			drawEdge(ge, null);
		}
		for (int i = 0; i < gn.inDegree(); i++) {
			GraphEdge ge = (GraphEdge)gn.getInEdge(i);
			if (ge.source != ge.target) {
				drawEdge(ge, getBackground());
				moveEdge(ge, gn, dx, dy);
				drawEdge(ge, null);
			}
		}
	}
	
	// return a location a little to the right of the center of line
	// connected a and b
	protected Point getLabelPosition(Point a, Point b) {
		Point p = new Point((a.x + b.x) / 2 + 50, (a.y + b.y) / 2);
		return p;
	}

	protected void moveEdge(GraphEdge ge, GraphNode source, int dx, int dy) {
		if (ge.source == ge.target) {
			// the calling method (moveAllEdges) is responsible for
			// ensuring not to call moveEdge twice for a loop edge.
			for (int i = 0; i < ge.points.size(); i++) {
				Point p = (Point)ge.points.elementAt(i);
				p.x += dx;
				p.y += dy;
			}
			if (ge.label != null && ge.label.location != null) {
				ge.label.location.x += dx;
				ge.label.location.y += dy;
			}
		} else {
			int n = ge.points.size();
			Point start_point, end_point;
			if (ge.source == source) {
				start_point = (Point)ge.points.elementAt(0);
				end_point = (Point)ge.points.elementAt(n - 1);
			} else {
				start_point = (Point)ge.points.elementAt(n - 1);
				end_point = (Point)ge.points.elementAt(0);
			}
			int width = start_point.x - end_point.x,
				height = start_point.y - end_point.y;
			for (int i = 0; i < n; i++) {
				Point p = (Point)ge.points.elementAt(i);
				if (width != 0)
					p.x += (dx * (p.x - end_point.x)) / width;
				else
					p.x += (dx * (p.y - end_point.y)) / height;
				if (height != 0)
					p.y += (dy * (p.y - end_point.y)) / height;
				else
					p.y += (dy * (p.x - end_point.x)) / width;
			}
			if (ge.label != null && ge.label.location != null) {
				ge.label.location.x += dx / 2;
				ge.label.location.y += dy / 2;

			}
		}
	}


	public Dimension getMinimumSize() {
		return new Dimension(150, 250);
	}
	public Dimension getPreferredSize() {
		return getSize();
	}

}
