package alexsa.gui.codeview;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Rectangle;
import java.awt.Point;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import javax.swing.JComponent;
import java.util.Enumeration;
import java.util.Vector;

import alexsa.data.tvp.*;
import alexsa.data.Tools;
import alexsa.data.Edge;
import alexsa.data.Node;
import alexsa.event.*;

public class CodeView extends JComponent implements MouseListener {

	protected Program
		program = null;
	
	protected Vector
		display_lines = null;

	protected int
		label_width = 0,
		code_width = 0,
		line_space = 0;

	protected boolean
		size_calculated = false;
	
	protected CfgNode
		current_node = null;

	public CodeView(Program p) {
		setFont(new Font("Monospaced", Font.PLAIN, 12));
		program = p;
		
		Node.dfs(p.start_node);
		
		// set the code associated with this edge to the result
		// of the title evaluation
		Vector edges = program.getCfgEdgeList();
		for (int i = 0; i < edges.size(); i++) {
			CfgEdge ce = (CfgEdge)edges.elementAt(i);
			ce.code = getTitleOfAction(ce.getAction()).trim();
		}
		
		program.resetCfgNodes();

		reconstructCode(p.start_node);
		
		createDisplayLines((CfgNode)p.start_node);
		
		cleanUp();
	}
	
	public void reconstructCode(Node n) {
		if (n.color != Node.WHITE)
			return;
		n.color = Node.GRAY;

		boolean either_or = eitherOrBranch(n);

		if (either_or) {
			Edge
				// incoming edge classified as back-edge
				back_in = getSingleBackInEdge(n),
				// outgoing edge classified as back-edge
				back_out = getSingleBackOutEdge(n);
			int inner = innerEdgeIndex(n);
		
			// test for while-loop
			if (back_in != null &&
				inner != -1 &&
				n.getOutEdge(inner).target.find(back_in.source))
			{
				n = new WhileLoop(n, program, inner, 1 - inner);
				Node exit_target = ((WhileLoop)n).goto_on_exit;
				boolean exit_was_white = exit_target.color == Node.WHITE;
				if (exit_was_white)
					exit_target.color = Node.GRAY;
				reconstructCode(((WhileLoop)n).while_condition.target);
				if (exit_was_white)
					exit_target.color = Node.WHITE;
				reconstructCode(exit_target);
			} else
			// test for do-while-loop
			if (back_in == null &&
				back_out != null &&
				back_out.target.outDegree() == 1 &&
				back_out.target.find(n))
			{
				n = new DoWhileLoop(n, program, (CfgEdge)back_out);
				reconstructCode(((DoWhileLoop)n).goto_on_exit.target);
			} else
			// test for if without else
			if (back_in == null &&
				inner != -1 &&
				n.getOutEdge(inner).target.find(n.getOutEdge(1 - inner).target))
			{
				n = new IfStatement(n, program, inner, 1 - inner);
				Node else_target = ((IfStatement)n).next_after_if_branch.target;
				boolean else_was_white = else_target.color == Node.WHITE;
				if (else_was_white)
					else_target.color = Node.GRAY;
				reconstructCode(((IfStatement)n).if_branch.target);
				if (else_was_white)
					else_target.color = Node.WHITE;
				reconstructCode(else_target);
			} else {
				CfgNode joining_point = (CfgNode)checkForSimpleIfElse(n, inner);
				// test for simple if/else
				if (inner != -1 && joining_point != null)
				{
					// simple if/else
					n = new SimpleIfElseStatement(n, program, inner, 1 - inner, joining_point);
					joining_point.color = Node.GRAY;
					reconstructCode(((SimpleIfElseStatement)n).if_branch.target);
					reconstructCode(((SimpleIfElseStatement)n).else_branch.target);
					joining_point.color = Node.WHITE;
					reconstructCode(joining_point);
				} else {
					// general if/else
					n = new GeneralIfElseStatement(n, program, inner, 1 - inner);
					reconstructCode(((GeneralIfElseStatement)n).if_branch.target);
					reconstructCode(((GeneralIfElseStatement)n).else_branch.target);
				}
			}
		} else if (n.outDegree() > 0) {
			for (int i = 0; i < n.outDegree(); i++) {
				Edge e = n.getOutEdge(i);
				reconstructCode(e.target);
			}
		}
		n.color = Node.BLACK;
	}

	private String getTitleOfAction(String s) {
		Action a = program.getAction(s.substring(0, s.indexOf('(')));
		String vars = s.substring(s.indexOf('(') + 1, s.lastIndexOf(')'));
		return a.printTitle(Tools.splitWithBraces(vars, ',', '[', ']'));
	}

	private Node checkForSimpleIfElse(Node n, int inner) {
		/* step one: check inner branch. The branch should
		   lead directly to some node n with exactly two incoming
		   edges. All nodes are supposed to be white (until n).
		   
		   Step two: check, whether else-branch leads directly to n.
		*/
		if (inner < 0)
			return null;
		
		Node end_of_if = n.getOutEdge(inner).target;
		while (end_of_if != null &&
			end_of_if.inDegree() == 1 &&
			end_of_if.outDegree() == 1 &&
			!(end_of_if instanceof ComplexNode) &&
			end_of_if.color == Node.WHITE)
		{
			end_of_if = end_of_if.getOutEdge(0).target;
		}

		// if we encounter a complex structure, we no longer have a simple
		// case. We also stop, if the supposed joining point has more than
		// two incoming edges.
		if (end_of_if == null ||
			end_of_if instanceof ComplexNode ||
			end_of_if.inDegree() > 2 ||
			end_of_if.color != Node.WHITE)
		{
			return null;
		}
		
		Node end_of_else = n.getOutEdge(1 - inner).target;
		while (end_of_else != null &&
			end_of_else.inDegree() == 1 &&
			end_of_else.outDegree() == 1 &&
			!(end_of_else instanceof ComplexNode) &&
			end_of_else.color == Node.WHITE)
		{
			end_of_else = end_of_else.getOutEdge(0).target;
		}

		if (end_of_else == end_of_if)
			return end_of_if;
		return null;
	}
	/* If there is exactly one edge classified as back edge entering n,
		return it.
		If there are no or multiple back edges entering n,
		return null
	*/
	public Edge getSingleBackInEdge(Node n) {
		int back_in_edge = -1;
		for (int i = n.inDegree() - 1; i >= 0; i--)
			if (n.getInEdge(i).type == Edge.BACK_EDGE) {
				if (back_in_edge == -1)
					back_in_edge = i;
				else
					return null;
			}
		if (back_in_edge != -1)
			return n.getInEdge(back_in_edge);
		else
			return null;
	}

	/* If there is exactly one edge classified as back edge leaving n,
		return it.
		If there are no or multiple back edges leaving n,
		return null
	*/
	public Edge getSingleBackOutEdge(Node n) {
		int back_out_edge = -1;
		for (int i = n.outDegree() - 1; i >= 0; i--)
			if (n.getOutEdge(i).type == Edge.BACK_EDGE) {
				if (back_out_edge == -1)
					back_out_edge = i;
				else
					return null;
			}
		if (back_out_edge != -1)
			return n.getOutEdge(back_out_edge);
		else
			return null;
	}

	/* returns true, if the given node has an either-or-branch, that is,
		a 2-place branch, where the actions do not change data but have
		inverted preconditions.
		See alexsa.data.tvp.Formula for inversion and equality of
		formulas.
		Example: an if-statement in tvp looks like this:
			n_1 x == null n_2
			n_1 x != null exit
		
		There is also the option to see two undefined actions as
		two either-or-conditions. This makes much nicer views whenever
		actions resembling data-comparisons are involved.
	*/ 
	public boolean eitherOrBranch(Node n) {
		if (n.outDegree() != 2)
			return false;
		Action
			action1 = program.getAction(((CfgEdge)n.getOutEdge(0)).identifier),
			action2 = program.getAction(((CfgEdge)n.getOutEdge(1)).identifier);

		if (alexsa.Preferences.summarize_unknown_branches &&
			action1 == action2 &&
			action1.new_formula == null &&
			action1.precondition == null &&
			action1.update_list == null)
		{
			return true;
		}
		return (action1.update_list == null && action2.update_list == null &&
			action1.new_formula == null && action2.new_formula == null &&
			action1.precondition != null && action2.precondition != null &&
			action1.precondition.invert().equals(action2.precondition));
	}
	
	/* of two edges one or none can be considered "inner".
		Given they both start at the same node, the inner one is
		the one leading to the next node in the order used in the tvp-file.
		The two given edges are not only supposed to start at the same
		node, but also have to appear adjoined in the tvp-file.
	*/
	public int innerEdgeIndex(Node n) {
		if (n.outDegree() != 2)
			return -1;
		Vector edges = program.getCfgEdgeList();
		int i1 = edges.indexOf(n.getOutEdge(0)),
			i2 = edges.indexOf(n.getOutEdge(1));
		int next_i;
		if (Math.abs(i1 - i2) != 1)
			next_i = Math.min(i1, i2) + 1;
		else
			next_i = Math.max(i1, i2) + 1;
		if (next_i >= edges.size())
			return -1;
		Edge next_edge = (Edge)edges.elementAt(next_i);
		if (n.getOutEdge(0).target == next_edge.source && n.getOutEdge(1).target != next_edge.source)
			return 0;
		else if (n.getOutEdge(0).target != next_edge.source && n.getOutEdge(1).target == next_edge.source)
			return 1;
		else
			return -1;
	}
	

	protected void createDisplayLines(CfgNode start) {
		program.resetCfgNodes();

		display_lines = new Vector();
		
		start.createDisplayLines(display_lines, 0);
		
		size_calculated = false;
		
	}
	
	/* cleanUp checks for every line, whether we need the "goto"-labels
	*/
	protected void cleanUp() {
		String next_label = null;
		DisplayLine next_line = null;
		
		for (int i = 0; i < display_lines.size(); i++) {
			DisplayLine dl = (DisplayLine)display_lines.elementAt(i);

			if (dl.target == null)
				continue;

			// let's take a look at the next line, including following detours.
			int next_index = dl.detour_index < 0 ? i + 1 : dl.detour_index;
			next_line = null;
			while (next_index >= 0 && next_index < display_lines.size()) {
				next_line = (DisplayLine)display_lines.elementAt(next_index);
				next_index = next_line.detour_index;
			}
			
			// two cases: either we reach implicitly or explicitly
			// the end of the code view while wanting to reach some 'exit'-state,
			// or we find some nice line with matching label for our goto.
			// in both cases, remove goto.
			if (next_line == null ||
				(next_line != null && next_index >= display_lines.size()))
			{
				CfgNode target = program.getCfgNode(dl.target);
				if (target.outDegree() == 0 && ! (target instanceof ComplexNode))
					dl.target = null;
			} else if (next_index < 0) {
				if (dl.target.equals(next_line.label))
					dl.target = null;
			}
		}
	}

	/* calcSize determines the size of the visual representation of the
		code view
	*/
	protected void calcSize(Graphics g) {
		FontMetrics fm = g.getFontMetrics(getFont());

		label_width = 0;
		code_width = 0;
		
		for (int i = 0; i < display_lines.size(); i++) {
			DisplayLine dl = (DisplayLine)display_lines.elementAt(i);
			if (dl.label != null)
				label_width = Math.max(label_width, fm.stringWidth(dl.label));
			
			if (dl.target != null)
				code_width = Math.max(code_width, fm.stringWidth(
					dl.code + "; goto " + dl.target +";"));
			else
				code_width = Math.max(code_width, fm.stringWidth(dl.code + ";"));
		}

		line_space = fm.getHeight();
		size_calculated = true;
	}


	public void paintComponent(Graphics g_old) {
		if (!size_calculated)
			calcSize(g_old);
	
		Graphics2D g = (Graphics2D)g_old;
		
		Font f = getFont();
		FontMetrics fm = g.getFontMetrics(f);
		g.setFont(f);
		int y = fm.getLeading() + fm.getAscent();

		for (int i = 0; i < display_lines.size(); i++) {
			DisplayLine dl = (DisplayLine)display_lines.elementAt(i);
			// if this line is supposed to be highlighted, put a nice blueish
			// bar under it
			if (active_lines.indexOf(dl) >= 0) {
				g.setColor(new Color(200, 200, 255));
				g.fillRect(0, y - fm.getLeading() - fm.getAscent(), getSize().width, line_space);
				g.setColor(Color.black);
			}
			g.drawString(dl.toString(), 5, y);
			y += line_space;
		}
	}
	
	public Vector getDisplayLines() {
		return display_lines;
	}
	
	public Dimension getPreferredSize() {
		return new Dimension(label_width + code_width + 20, display_lines.size() * line_space);
	}
	
	public Dimension getMinimumSize() {
		return getPreferredSize();
	}

	public CfgNode getCurrentNode() {
		return current_node;
	}
	
	protected Vector
		active_lines = new Vector();

	public void activate(CfgNode n) {
		if (active_lines.size() > 0)
			active_lines.removeAllElements();

		for (int i = display_lines.size() - 1; i >= 0; i--) {
			DisplayLine dl = (DisplayLine)display_lines.elementAt(i);
			if (dl.cfg_edge != null && dl.cfg_edge.source == n)
				active_lines.addElement(dl);
		}
		current_node = n;
		repaint(0);
	}
	public Object getActiveObject() {
		if (active_lines.size() == 0)
			return null;
		else if (active_lines.size() == 1)
			return active_lines.elementAt(0);
		else
			return active_lines;
	}
	public void deactivate() {
		if (active_lines.size() > 0) {
			active_lines.removeAllElements();
			repaint(0);
		}
	}

	Vector graph_listeners;
	public void addGraphListener(GraphListener gl) {
		if (graph_listeners == null)
			graph_listeners = new Vector();
		if (graph_listeners.indexOf(gl) < 0)
			graph_listeners.addElement(gl);
		if (graph_listeners.size() == 1)
			addMouseListener(this);
	}
	public void removeGraphListener(GraphListener gl) {
		if (graph_listeners == null)
			return;
		graph_listeners.removeElement(gl);
		if (graph_listeners.size() == 0) {
			graph_listeners = null;
			removeMouseListener(this);
		}
	}
	public void deliverGraphEvent(Object o) {
		if (graph_listeners != null) {
			GraphEvent ge = new GraphEvent(this, o);
			for (int i = 0; i < graph_listeners.size(); i++)
				((GraphListener)graph_listeners.elementAt(i)).graphAction(ge);
		}
	}

	protected DisplayLine getDisplayLineAt(int x, int y) {
		if (line_space == 0)
			return null;
		y = y / line_space;
		if (y >= 0 && y < display_lines.size())
			return (DisplayLine)display_lines.elementAt(y);
		return null;
	}

    public void mouseClicked(MouseEvent e) {
		Object o = null;
		DisplayLine dl = getDisplayLineAt(e.getX(), e.getY());
		if (dl == null || dl.cfg_edge == null)
			return;
		
		if (e.getClickCount() == 1)
			deliverGraphEvent(dl.cfg_edge.source);
		else if (e.getClickCount() == 2)
			deliverGraphEvent(dl.cfg_edge);
	}
    public void mousePressed(MouseEvent e) {}
    public void mouseReleased(MouseEvent e) {}
    public void mouseEntered(MouseEvent e) {}
    public void mouseExited(MouseEvent e) {}

}
